@@ -17,9 +17,9 @@ __author__ = "Felipe Torres González"
 __copyright__ = "Copyright 2021, ESS MCH Tools"
 __credits__ = ["Felipe Torres González", "Ross Elliot", "Jeong Han Lee"]
 __license__ = "GPL-3.0"
-__version__ = "0.1"
-__maintainer__ = "Felipe Torres González"
-__email__ = "felipe.torresgonzalez@ess.eu"
+__version__ = "1.1"
+__maintainer__ = "Ross Elliot"
+__email__ = "ross.elliot@ess.eu"
 __status__ = "Development"
 
 
@@ -60,7 +60,7 @@ class ConnTimeout(Exception):
 
 
 class NoRouteToDevice(Exception):
-    """Connection timeout Exception."""
+    """No route to device Exception."""
 
     def __init__(self, *args):
         if args:
@@ -75,6 +75,22 @@ class NoRouteToDevice(Exception):
             return "NoRouteToDevice has been raised"
 
 
+class ConnectionRefused(Exception):
+    """Connection refused Exception."""
+
+    def __init__(self, *args):
+        if args:
+            self.message = args[0]
+        else:
+            self.message = None
+
+    def __str__(self):
+        if self.message:
+            return "ConnectionRefused, {0} ".format(self.message)
+        else:
+            return "ConnectionRefused has been raised"
+
+
 class FeatureNotSupported(Exception):
     """Feature not supported exception.
 
