@@ -20,16 +20,16 @@ Main features supported by the API:
 from __future__ import annotations
 import abc
 import enum
-import logging
+from logging import Logger
 from collections import OrderedDict
 
 __author__ = "Felipe Torres González"
 __copyright__ = "Copyright 2021, ESS MCH Tools"
 __credits__ = ["Felipe Torres González", "Ross Elliot", "Jeong Han Lee"]
 __license__ = "GPL-3.0"
-__version__ = "1.0"
-__maintainer__ = "Felipe Torres González"
-__email__ = "felipe.torresgonzalez@ess.eu"
+__version__ = "1.1"
+__maintainer__ = "Ross Elliot"
+__email__ = "ross.elliot@ess.eu"
 __status__ = "Development"
 
 
@@ -73,7 +73,8 @@ class GenDevInterface(metaclass=abc.ABCMeta):
         hostname: str = None,
         vlan: str = None,
         mac_address: str = None,
-        logger: logging.Logger = None,
+        logger: Logger = None,
+        log_en: bool = False,
     ):
         """Class constructor.
 
@@ -233,3 +234,18 @@ class GenDevInterface(metaclass=abc.ABCMeta):
             ConnectionError: If the device is not accessible.
         """
         raise NotImplementedError
+
+    @abc.abstractmethod
+    def reboot_slot(self, amc_num) -> tuple[bool, str]:
+        """ Method to power cycle an individual AMC slot.
+
+        Args:
+            amc_num: AMC slot number to power cycle.
+
+        Returns:
+            A tuple containing the a success flag in the first position. In the \
+            second position:
+            - On success, an empty string.
+            - On failure, a string containing the failure message.
+        """
+        raise NotImplementedError
