@@ -15,10 +15,13 @@ is achieving a reliable solution with the best performance.
 
 from __future__ import annotations
 import logging
+from pprint import pformat
 from enum import IntEnum
 from collections import OrderedDict
+from difflib import Differ
 from ..gendev_interface import GenDevInterface, ConnType
 from ..gendev_err import ConnNotImplemented, FeatureNotSupported, DHCPEnableFailed
+from ..gendev_logger import GenDevLogger
 from .nat_mch_web import NATMCHWeb
 from .nat_mch_telnet import NATMCHTelnet
 from .nat_mch_moxa import NATMCHMoxa
@@ -27,9 +30,9 @@ __author__ = "Felipe Torres González"
 __copyright__ = "Copyright 2021, ESS MCH Tools"
 __credits__ = ["Felipe Torres González", "Ross Elliot", "Jeong Han Lee"]
 __license__ = "GPL-3.0"
-__version__ = "1.0"
-__maintainer__ = "Felipe Torres González"
-__email__ = "felipe.torresgonzalez@ess.eu"
+__version__ = "1.1"
+__maintainer__ = "Ross Elliot"
+__email__ = "ross.elliot@ess.eu"
 __status__ = "Development"
 
 
@@ -79,6 +82,7 @@ class NATMCH(GenDevInterface):
         conn_ip_address: str = None,
         conn_port: int = None,
         logger: logging.Logger = None,
+        log_to_file: bool = False,
     ):
         """Class constructor.
 
@@ -110,13 +114,28 @@ class NATMCH(GenDevInterface):
         self.hostname = ""
         self.mac_address = ""
         self.backplane = backplane
-        self.logger = logger
         self._conn_ip_address = conn_ip_address
         self._conn_port = conn_port
+        self._log_to_file = log_to_file
+
+        # Create a GenDevLogger instance from Logger object
+        self.logger = GenDevLogger(logger=logger, log_to_file=self._log_to_file)
+        self.logger.logger.info(
+            "GenDev::Constructor - A new device has been registered"
+            "\tDevice model: {}".format(self.device_model)
+        )
+
+        if log_to_file:
+            self.logger.logger.info(
+                "Log file located at: {}".format(self.logger.logname)
+            )
 
         # Open the valid connections
         if ConnType.ETHER in self.allowed_conn:
-            self._eth_conn = NATMCHWeb(self.ip_address)
+            self._eth_conn = NATMCHWeb(
+                self.ip_address,
+                logger=self.logger,
+            )
         if ConnType.TELNET in self.allowed_conn:
             self._tel_conn = NATMCHTelnet(
                 ip_address=self.ip_address,
@@ -133,7 +152,7 @@ class NATMCH(GenDevInterface):
                 moxa_ip_address=self._conn_ip_address,
                 port=self._conn_port,
                 backplane=self.backplane,
-                logger=None,
+                logger=self.logger,
                 hostname=self.hostname,
             )
         if ConnType.SSH in self.allowed_conn:
@@ -141,13 +160,6 @@ class NATMCH(GenDevInterface):
                 "The SSH interface is not implemented" " for NAT MCHs."
             )
 
-        # TBD: support logging
-        # if self.logger is not None:
-        #     self.logger.info(
-        #         'GenDev::Constructor - A new device has been registered'
-        #         '\tDevice model: {}'.format(self.device_model)
-        #         )
-
         # This dict holds a look-up table to translate argument keys, given to the
         # methods of this class, to keys from the configuration files.
         self._cfgkeys = {
@@ -202,16 +214,42 @@ class NATMCH(GenDevInterface):
         elif ConnType.MOXA in self.allowed_conn:
             valid, response = self._mox_conn.device_info()
         else:
-            raise FeatureNotSupported(
-                "Impossible to retrieve the device"
-                "information with the given allowed"
+            msg = (
+                "Impossible to retrieve the device "
+                "information with the given allowed "
                 "communication interfaces to the MCH."
             )
+            self.logger.logger.error(msg)
+            raise FeatureNotSupported(msg)
 
         if valid:
             self.mac_address = response["Network"]["mac_address"]
             self.serial_num = response["Board"]["serial_num"]
             self.fw_ver = response["Board"]["fw_ver"]
+            self.logger.logger.info(
+                "\n[Board info]\n \
+                   Serial number:    {}\n \
+                   FPGA version:     {}\n \
+                   Firmware version: {}\n \
+                   MCU Version:      {}\n".format(
+                    self.serial_num,
+                    response["Board"]["fpga_ver"],
+                    response["Board"]["fw_ver"],
+                    response["Board"]["mcu_ver"],
+                )
+            )
+            self.logger.logger.info(
+                "\n[Network info]\n \
+                   IP Address:  {}\n \
+                   MAC Address: {}\n \
+                   Subnet:      {}\n \
+                   Gateway:     {}\n".format(
+                    self.ip_address,
+                    self.mac_address,
+                    response["Network"]["subnet_address"],
+                    response["Network"]["gateway_address"],
+                )
+            )
 
         return valid, response
 
@@ -240,14 +278,15 @@ class NATMCH(GenDevInterface):
         elif ConnType.MOXA in self.allowed_conn:
             success, response = self._mox_conn.set_dhcp_mode()
         else:
-            raise FeatureNotSupported(
-                (
-                    "Setting the DHCP mode is not supported for the given"
-                    "connection type"
-                )
+            msg = (
+                "Setting the DHCP mode is not supported for the given "
+                "connection type"
             )
+            self.logger.logger.error(msg)
+            raise FeatureNotSupported(msg)
 
         if not success:
+            self.logger.logger.error(response)
             raise DHCPEnableFailed(response)
 
         return success, response
@@ -280,11 +319,19 @@ class NATMCH(GenDevInterface):
         elif ConnType.MOXA in self.allowed_conn:
             valid, response = self._mox_conn.update_fw(fw_version)
         else:
-            raise FeatureNotSupported(
-                "Impossible to update the fw of the"
-                " device with the given allowed"
+            msg = (
+                "Impossible to update the fw of the "
+                "device with the given allowed "
                 "communication interfaces to the MCH."
             )
+            self.logger.logger.error(msg)
+            raise FeatureNotSupported(msg)
+
+        if valid:
+            self.logger.logger.info(response)
+        else:
+            self.logger.logger.error(response)
+
         return valid, response
 
     def set_configuration(
@@ -328,10 +375,15 @@ class NATMCH(GenDevInterface):
         elif ConnType.MOXA in self.allowed_conn and category == "backplane":
             valid, response = self._mox_conn.set_configuration(category, data, apply)
         else:
-            raise FeatureNotSupported(
-                "Retrieving the MCH configuration using the"
-                " {} backend is not implemented.".format(str(self.allowed_conn))
+            msg = (
+                "Setting the MCH configuration using the "
+                "{} backend is not implemented.".format(str(self.allowed_conn))
             )
+            self.logger.logger.error(msg)
+            raise FeatureNotSupported(msg)
+
+        if not valid:
+            self.logger.logger.error(response)
 
         return valid, response
 
@@ -366,10 +418,17 @@ class NATMCH(GenDevInterface):
         if ConnType.ETHER in self.allowed_conn:
             valid, response = self._eth_conn.get_configuration(category)
         else:
-            raise FeatureNotSupported(
-                "Retrieving the MCH configuration using the"
-                " {} backend is not implemented.".format(str(self.allowed_conn))
+            msg = (
+                "Retrieving the MCH configuration using the "
+                "{} backend is not implemented.".format(str(self.allowed_conn))
             )
+            self.logger.logger.error(msg)
+            raise FeatureNotSupported(msg)
+
+        if valid:
+            self.logger.logger.debug(pformat(response))
+        else:
+            self.logger.logger.error(response)
 
         return valid, response
 
@@ -396,47 +455,97 @@ class NATMCH(GenDevInterface):
             checking was unsuccesful.
         """
         if ConnType.ETHER not in self.allowed_conn:
-            raise FeatureNotSupported(
+            msg = (
                 "Accessing the MCH configuration is only possible "
                 "when using the Web backend."
             )
+            self.logger.logger.error(msg)
+            raise FeatureNotSupported(msg)
 
         # Is `config` good?
         try:
             valid_args = self._eth_conn._check_arguments(config, category)
             if not valid_args[0]:
+                self.logger.logger.info(
+                    "Configuration check for {} passed successfully.".format(category)
+                )
                 return valid_args
         except KeyError:
-            return (False, "The given configuration set contains unknown parameters.")
+            msg = "The given configuration set contains unknown parameters."
+            self.logger.logger.error(msg)
+            return (False, msg)
 
         valid, mch_config = self._eth_conn.get_configuration(category)
+
         if not valid:
+            self.logger.logger.info(
+                "Configuration check for {} failed.".format(category)
+            )
             return valid, mch_config
+
         key = self._cfgkeys[category]["cfgkey"]
 
         # The backplane configuration is a special thing
         if category == "backplane":
-            return mch_config[key]["file"] == config[key]["file"], ""
+            # The initial comment lines are stripped from the golden config
+            # when it is applied to the MCH.
+            #
+            # Strip the first 11 lines before comparing
+
+            # Split string by newline character
+            mch_golden_conf_list = config[key]["file"].splitlines(keepends=True)
+            mch_conf_list = mch_config[key]["file"].splitlines(keepends=True)
+
+            # Create Differ() object
+            differ = Differ()
+            diff = list(differ.compare(mch_conf_list, mch_golden_conf_list))
+
+            if mch_config[key]["file"] == config[key]["file"]:
+                ret = (True, "")
+            else:
+                ret = (
+                    False,
+                    "Backplane configuration setting failed. See log for details.",
+                )
+                self.logger.logger.debug(
+                    "Backplane configuration does not match golden config. Diff provided:"
+                )
+                self.logger.logger.debug(pformat(diff))
+
+            return ret
 
         # Fast comparison: keys
         if mch_config[key].keys() != config[key].keys():
-            return (
-                False,
-                "The given set of parameters doesn't match the values from the MCH.",
-            )
+            msg = "The given set of parameters doesn't match the values from the MCH."
+            self.logger.logger.error(msg)
+            return (False, msg)
 
         # Second, compare the content
         result = (True, "")
+        skip_keys = []
+        if category == "basecfg":
+            skip_keys = ["DHCP parameter"]
         try:
-            result = self._parse_config(mch_config, config, key)
+            result = self._parse_config(mch_config, config, key, skip_keys)
         except KeyError as e:
             # Add something to the log
-            result = (False, "The configuration is missing a key: {}".format(e))
+            msg = "The configuration is missing a key: {}".format(e)
+            self.logger.logger.error(msg)
+            result = (False, msg)
+
+        if not result[0]:
+            self.logger.logger.error(
+                "Configuration check for {} failed.".format(category)
+            )
+        else:
+            self.logger.logger.info(
+                "Configuration check for {} succeeded.".format(category)
+            )
 
         return result
 
     def _parse_config(
-        self, dict1: OrderedDict, dict2: OrderedDict, key, logger=None
+        self, dict1: OrderedDict, dict2: OrderedDict, key, skip_keys: list = []
     ) -> tuple[bool, str]:
         """Compare two dictionaries, item by item.
 
@@ -456,8 +565,7 @@ class NATMCH(GenDevInterface):
             dict2 (OrderedDict): Slave dictionary. This one is accessed using the keys from \
             `dict1`. If using untrusted content, like coming from an user, introduce it using \
             this argument.
-            logger (Logger): Logger instance. Defaults to None.
-
+            skip_keys (list): List of keys to skip during checks.
         Returns:
             tuple[bool, str]: Success status and error message when a difference was found.
 
@@ -467,27 +575,111 @@ class NATMCH(GenDevInterface):
         """
         for k in dict1[key]:
             if isinstance(dict1[key][k], OrderedDict):
-                return self._parse_config(dict1[key], dict2[key], k)
+                if k in skip_keys:
+                    self.logger.logger.debug('Skipping check for "{}"'.format(k))
+                    continue
+                else:
+                    self.logger.logger.debug("{}:".format(k))
+                    check, msg = self._parse_config(dict1[key], dict2[key], k)
+                    if not check:
+                        return (check, msg)
             elif dict1[key][k] != dict2[key][k]:
-                # TODO: add this to a log
-                print("{} != {}".format(dict1[key][k], dict2[key][k]))
-                return (False, "The configuration check for {} failed.".format(k))
+                # DHCP hostname cannot be checked, as it differs for each
+                # MCH
+                if k not in skip_keys:
+                    self.logger.logger.error(
+                        "{} != {}".format(dict1[key][k], dict2[key][k])
+                    )
+                    msg = "The configuration check for {} failed.".format(k)
+                    self.logger.logger.error(msg)
+                    return (False, msg)
+            else:
+                self.logger.logger.debug(
+                    "-- {}, {} == {}".format(k, dict1[key][k], dict2[key][k])
+                )
 
         return (True, "")
 
-    def _reboot(self, sleep: int = 50):
+    def _reboot(self, sleep: int = 60):
         """Internal method to reboot the MCH after a timeout.
 
         Args:
             sleep: Number of seconds to wait after rebooting the device.
         """
-        if ConnType.TELNET in self.allowed_conn:
+        if ConnType.ETHER in self.allowed_conn:
+            self._eth_conn._reboot(sleep)
+        elif ConnType.TELNET in self.allowed_conn:
             self._tel_conn._reboot(sleep)
-        if ConnType.MOXA in self.allowed_conn:
-            self._moxa_conn._reboot(sleep)
+        elif ConnType.MOXA in self.allowed_conn:
+            self._mox_conn._reboot(sleep)
         else:
-            raise FeatureNotSupported(
-                "Impossible to reboot the device"
-                "with the given allowed"
+            msg = (
+                "Impossible to reboot the device "
+                "with the given allowed "
                 "communication interfaces to the MCH."
             )
+            self.logger.logger.error(msg)
+            raise FeatureNotSupported(msg)
+
+    def reboot_slot(self, amc_num) -> tuple[bool, str]:
+        """ Method to power cycle an individual AMC slot.
+
+        Args:
+            amc_num: AMC slot number to power cycle.
+
+        Returns:
+            A tuple containing the a success flag in the first position. In the \
+            second position:
+            - On success, an empty string.
+            - On failure, a string containing the failure message.
+        """
+        if ConnType.TELNET in self.allowed_conn:
+            success, response = self._tel_conn.reboot_slot(amc_num)
+        elif ConnType.MOXA in self.allowed_conn:
+            success, response = self._mox_conn.reboot_slot(amc_num)
+        else:
+            msg = (
+                "The current connection to the MCH "
+                "does not allow for power cycling "
+                "of individual slots."
+            )
+            self.logger.logger.error(msg)
+            raise FeatureNotSupported(msg)
+
+        if not success:
+            self.logger.logger.error(response)
+        else:
+            self.logger.logger.info(
+                "Successfully rebooted AMC in slot {}.".format(amc_num)
+            )
+
+        return (success, response)
+
+    def reboot_all_slots(self) -> tuple[bool, str]:
+        """ Method to power cycle all AMC slots.
+
+        Returns:
+            A tuple containing the a success flag in the first position. In the \
+            second position:
+            - On success, an empty string.
+            - On failure, a string containing the failure message.
+        """
+        if ConnType.TELNET in self.allowed_conn:
+            success, response = self._tel_conn.reboot_all_slots()
+        elif ConnType.MOXA in self.allowed_conn:
+            success, response = self._mox_conn.reboot_all_slots()
+        else:
+            msg = (
+                "The current connection to the MCH "
+                "does not allow for power cycling "
+                "of slots."
+            )
+            self.logger.logger.error(msg)
+            raise FeatureNotSupported(msg)
+
+        if not success:
+            self.logger.logger.error(response)
+        else:
+            self.logger.logger.info("Successfully rebooted all AMC slots.")
+
+        return (success, response)
