@@ -12,13 +12,13 @@ class.
 import time
 import re
 from gendev_tools.nat_mch.nat_mch_telnet import NATMCHTelnet
-from logging import Logger
+from ..gendev_logger import GenDevLogger
 
 __author__ = "Ross Elliot"
 __copyright__ = "Copyright 2021, ESS MTCA Tools"
 __credits__ = ["Felipe Torres González", "Jeong Han Lee"]
 __license__ = "GPL-3.0"
-__version__ = "0.1"
+__version__ = "1.1"
 __maintainer__ = "Ross Elliot"
 __email__ = "ross.elliot@ess.eu"
 __status__ = "Development"
@@ -44,8 +44,9 @@ class NATMCHMoxa(NATMCHTelnet):
         moxa_ip_address: str,
         port: int,
         backplane,
-        logger: Logger = None,
+        logger: GenDevLogger = None,
         hostname: str = None,
+        **kwargs,
     ):
         """Class constructor.
 
@@ -67,6 +68,11 @@ class NATMCHMoxa(NATMCHTelnet):
         self.port = 4000 + port
         self.moxa_ip_address = moxa_ip_address
 
+        if "log_to_file" in kwargs:
+            log_to_file = kwargs["log_to_file"]
+        else:
+            log_to_file = False
+
         # Initialise parent class and inherit methods and properties
         super().__init__(
             ip_address=mch_ip_address,
@@ -75,6 +81,7 @@ class NATMCHMoxa(NATMCHTelnet):
             logger=logger,
             hostname=hostname,
             telnet_ip_address=moxa_ip_address,
+            log_to_file=log_to_file,
         )
 
     def _send_command(
