@@ -22,17 +22,22 @@ import time
 import socket
 import ipaddress
 from collections import OrderedDict
-from ..gendev_err import ConnTimeout, NoRouteToDevice, FeatureNotSupported
+from ..gendev_err import (
+    ConnTimeout,
+    NoRouteToDevice,
+    FeatureNotSupported,
+    ConnectionRefused,
+)
+from ..gendev_logger import GenDevLogger
 from telnetlib import Telnet
-from logging import Logger
 
 __author__ = ["Felipe Torres González", "Ross Elliot"]
 __copyright__ = "Copyright 2021, ESS MTCA Tools"
 __credits__ = ["Jeong Han Lee"]
 __license__ = "GPL-3.0"
-__version__ = "1.0"
-__maintainer__ = "Felipe Torres González"
-__email__ = "felipe.torresgonzalez@ess.eu"
+__version__ = "1.1"
+__maintainer__ = "Ross ELliot"
+__email__ = "ross.elliot@ess.eu"
 __status__ = "Development"
 
 
@@ -60,7 +65,7 @@ class NATMCHTelnet:
         ip_address: str,
         backplane,
         port: int = 23,
-        logger: Logger = None,
+        logger: GenDevLogger = None,
         hostname: str = None,
         open: bool = True,
         **kwargs,
@@ -72,6 +77,8 @@ class NATMCHTelnet:
             backplane(BackplaneType): type of backplane in which the MCH is installed.
             port(int): port of the Telnet service (usually, 23).
             logger(Logger): reference to a logger that is being used.
+            log_to_file(Bool): boolean flag to indicate if the log should be written
+            to a file.
             hostname(str): hostname of the MCH.
             open(bool): whether to establish the connection at instantiation time or later. \
             Set this argument to False when a deferred link establishment is aimed.
@@ -80,7 +87,7 @@ class NATMCHTelnet:
             connecting via a MOXA (optional).
             **tftp_server_address(str): IP address of the machine hosting the TFTP \
             server with the fw images. Default value = **172.30.4.69**.
-            **tftp_server_fw_path(str): path to the fw binaries in the TFPT server. \
+            **tftp_server_fw_path(str): path to the fw binaries in the TFTP server. \
             Default value = "fw/".
 
         Raises:
@@ -110,6 +117,22 @@ class NATMCHTelnet:
         else:
             self._fw_path = "fw/"
 
+        if "log_to_file" in kwargs:
+            self._log_to_file = kwargs["log_to_file"]
+        else:
+            self._log_to_file = False
+
+        # Initialise Logger instance
+        if isinstance(logger, (GenDevLogger)):
+            self.logger = logger
+        else:
+            self.logger = GenDevLogger(logger=logger, log_to_file=self._log_to_file)
+
+        self.logger.logger.info(
+            "NATMCHTelnet - NAT MCH Telnet instance created."
+            "\tMCH IP Address: {}".format(self.ip_address)
+        )
+
         if open:
             self.open()
             self._session_open = True
@@ -132,31 +155,52 @@ class NATMCHTelnet:
         self._match_dhcp_state = re.compile("dhcp state +: (.+?(?=\r))")
 
     def open(self):
-        """Stablish the connection to the target MCH.
+        """Establish the connection to the target MCH.
 
         Raises:
             ConnTimeout: if the connection to the device gets lost.
             NoRouteToDevice: if the device is not reachable.
+            ConnectionRefused: if the telnet connection is refused.
         """
+        self.logger.logger.debug(
+            "Opening Telnet connection to {}:{}.".format(
+                self._telnet_ip_address, self.port
+            )
+        )
         try:
             self._session = Telnet(self._telnet_ip_address, self.port, timeout=10)
             self._session_open = True
         except Exception as e:
             if isinstance(e, socket.timeout):
-                raise ConnTimeout(
-                    "Timeout while opening the link to the MCH using Telnet"
+                msg = "Timeout while opening the link to the MCH using Telnet."
+                self.logger.logger.error(msg)
+                raise ConnTimeout(msg)
+            elif isinstance(e, OSError) and (e.errno == 113 or e.errno == 101):
+                msg = "Check the connectivity to the MCH" " using the IP: {}.".format(
+                    self._telnet_ip_address
                 )
-            elif isinstance(e, OSError) and e.errno == 113:
-                raise NoRouteToDevice(
-                    "Check the connectivity to the MCH"
-                    " using the IP: {}".format(self._telnet_ip_address)
+                self.logger.logger.error(msg)
+                raise NoRouteToDevice(msg)
+            elif isinstance(e, OSError) and (e.errno == 111):
+                msg = (
+                    "Cannot connect to Telnet port {}:{} "
+                    "Please close any existing "
+                    "connections.".format(self._telnet_ip_address, self.port)
                 )
+                self.logger.logger.error(msg)
+                raise ConnectionRefused(msg)
+            else:
+                self.logger.logger.error("Unhandled exception: {}".format(e.strerror))
+                raise e
 
     def close(self):
         """Close the Telnet session.
 
         This is needed when multiple methods of this module are called in a row.
         """
+        self.logger.logger.debug(
+            "Closing Telnet connection to {}:{}.".format(self.ip_address, self.port)
+        )
         self._session.close()
         self._session_open = False
 
@@ -202,8 +246,12 @@ class NATMCHTelnet:
             sleep: indicates how many seconds to wait after returning from the
             method. Write a 0 to avoid it.
         """
+        self.logger.logger.info("Rebooting MCH device.")
+        if self._session_open is False:
+            self.open()
         self._send_command("reboot")
-        self._session_open = False
+        self.close()
+        # self._session_open = False
         time.sleep(sleep)
 
     def _read_command(self) -> str:
@@ -225,6 +273,11 @@ class NATMCHTelnet:
             True on success.
             False on failure.
         """
+        self.logger.logger.info("Setting hostname to {}".format(self.hostname))
+
+        if self._session_open is False:
+            self.open()
+
         self._send_command("mchcfg")
         # Modify
         self._send_command("11", clear_buffer=False)
@@ -238,6 +291,8 @@ class NATMCHTelnet:
         # success = self._check_hostname()
         success = True
 
+        self.close()
+
         return success
 
     def _check_hostname(self) -> bool:
@@ -249,6 +304,9 @@ class NATMCHTelnet:
             False if the hostname setting does not match the expected string.
         """
 
+        if self._session_open is False:
+            self.open()
+
         self._send_command("ni")
         network_info = self._read_command()
 
@@ -258,6 +316,8 @@ class NATMCHTelnet:
         if mch_hostname == self.hostname:
             success = True
 
+        self.close()
+
         return success
 
     def _enable_dhcp(self):
@@ -270,10 +330,10 @@ class NATMCHTelnet:
 
         self._send_command("mchcfg")
         self._send_command("3", clear_buffer=False)
-        for _ in range(0, 4):
+        for _ in range(0, 5):
             self._send_command("", clear_buffer=False)
         self._send_command("2", clear_buffer=False)
-        for _ in range(0, 8):
+        for _ in range(0, 7):
             self._send_command("", clear_buffer=False)
         self._send_command("q", clear_buffer=False)
 
@@ -296,6 +356,9 @@ class NATMCHTelnet:
             False if DHCP is disabled
         """
 
+        if self._session_open is False:
+            self.open()
+
         self._send_command("ni")
         network_info = self._read_command()
 
@@ -304,6 +367,8 @@ class NATMCHTelnet:
         if dhcp_state == " enabled":
             enabled = True
 
+        self.close()
+
         return enabled
 
     def _clear_ip_address(self):
@@ -317,6 +382,10 @@ class NATMCHTelnet:
 
     def _set_ip_addr(self):
         """Set the IP address of the MCH device."""
+
+        if self._session_open is False:
+            self.open()
+
         # Get network information
         self._send_command("ni")
         network_info = self._read_command()
@@ -348,11 +417,14 @@ class NATMCHTelnet:
         # Check we are at the confirmation prompt
         response = self._read_command()
         if not response.endswith("Are you really sure ?"):
+            self.close()
             return False  # Failure
 
         # Issue confirmation 'y'
         self._send_command("y", clear_buffer=False)
 
+        self.close()
+
         # Success
         return True
 
@@ -455,6 +527,8 @@ class NATMCHTelnet:
             NoValidConn: If no valid connection types supporting this feature
                          are used by the device.
         """
+        self.logger.logger.info("Enabling DHCP mode.")
+
         if self._session_open is False:
             self.open()
 
@@ -502,6 +576,8 @@ class NATMCHTelnet:
             about the failure.
             If success, it returns True,
         """
+        self.logger.logger.info("Updating MCH firmware")
+
         if self._session_open is False:
             self.open()
 
@@ -612,12 +688,13 @@ class NATMCHTelnet:
                 "{}u".format(int(self.backplane)),
                 option,
             ),
-            sleep=6,
+            sleep=3,
             clear_buffer=False,
         )
         # Read the buffer to check whether the file was found in the TFTP server
         validfile = self._read_command()
-        if "ERR" in validfile:
+        if "TFTP: getting file done" not in validfile:
+            self.close()
             return False, "The backplane file wasn't found in the server. See the log."
 
         self._send_command("y", clear_buffer=False)
@@ -626,7 +703,389 @@ class NATMCHTelnet:
         if apply:
             self._reboot()
 
+        self.close()
+
         return (
             True,
             "Expect some delay until the MCH is accessible in the network again.",
         )
+
+    def _get_slot_state(self, amc_num, close_conn: bool = True):
+        """Internal method to get the current state of an
+        AMC slot.
+
+        Args:
+            amc_num(int): AMC slot number.
+            close_conn(bool): Whether to close the telnet
+               session when returning, or not.
+               Default: True
+
+        Returns:
+            A string containing the state code {M1, M2, ... M7}
+        """
+        if self._session_open is False:
+            self.open()
+
+        # Get fru_info for all AMCs
+        self._send_command("show_fru")
+        fru_info = self._read_command()
+
+        # Parse fru_info for line containing the requested AMC, and parse
+        # the state
+        match = re.search(r"AMC{}(\s+)(M\d)".format(amc_num), fru_info)
+
+        # Check if the Telnet session should be closed
+        if close_conn is True:
+            self.close()
+
+        if match is None:
+            success = False
+            state = "Unable to determine state of AMC {}. " "Check AMC number.".format(
+                amc_num
+            )
+        else:
+            success = True
+            state = match.group(2)
+
+        # State is in group 2 of returned regex search
+        return (success, state)
+
+    def _start_slot(self, amc_num, sleep: int = 2, upstream=False):
+        """Internal method to start the card in an AMC slot
+
+        Args:
+            amc_num(int): AMC slot number to power on.
+            sleep(int): Time to allow for the AMC to power down.
+                        Default: 2 seconds
+
+        Returns:
+            True if successful
+            False if unsuccessful
+        """
+        success = False
+
+        # MCH fru_start command expects fru number, instead of AMC
+        # number, where:
+        #     fru_num = amc_num + 4
+        fru_num = amc_num + 4
+
+        # Check current state of the AMC, and leave Telnet session
+        # open
+        res, state = self._get_slot_state(amc_num, False)
+
+        # If unable to get state of slot, return
+        if res is False:
+            self.close()
+            return False
+
+        if state == "M4":
+            self.logger.logger.info("AMC{} is already powered up.".format(amc_num))
+            success = True
+        else:
+            # Sleep to allow previous telnet connection to properly
+            # close
+            time.sleep(1)
+
+            if self._session_open is False:
+                self.open()
+
+            self.logger.logger.info("Powering up AMC {}.".format(amc_num))
+            # Send fru_start command
+            self._send_command("fru_start {}".format(fru_num))
+            fru_start = self._read_command()
+
+            # Check if the slot has a start up delay
+            if upstream or ("Upstream Power Up Delay active!" in fru_start):
+                self.logger.logger.info(
+                    "Slot has upstream delay set. Waiting 60s for AMC to power on..."
+                )
+                sleep = 65
+
+            # Sleep to allow device in AMC slot to power on
+            time.sleep(sleep)
+
+            # Get updated state
+            _, state = self._get_slot_state(amc_num)
+
+            if state == "M4":
+                success = True
+                self.logger.logger.info(
+                    "AMC {} powered up successfully.".format(amc_num)
+                )
+            else:
+                self.logger.logger.warning(
+                    "Failed to power on slot. Current state: {}.".format(state)
+                )
+
+        self.close()
+
+        return success
+
+    def _stop_slot(self, amc_num, sleep: int = 2):
+        """Internal method to stop the card in an AMC slot
+
+        Args:
+            amc_num(int): AMC slot number to power off.
+            sleep(int): Time to allow for the AMC to power down.
+                        Default: 2 seconds
+
+        Returns:
+            True if successful
+            False if unsuccessful
+        """
+        success = False
+
+        # MCH shutdown command expects fru number, instead of AMC
+        # number, where:
+        #     fru_num = amc_num + 4
+        fru_num = amc_num + 4
+
+        # Check current state of the AMC, and leave Telent
+        # session open
+        res, state = self._get_slot_state(amc_num, False)
+        # If unable to get state of slot, return
+        if res is False:
+            self.logger.logger.warning(
+                "Unable to determine state of slot {}:\n  {}".format(amc_num, state)
+            )
+            self.close()
+            return False
+
+        if state == "M1":
+            self.logger.logger.info("AMC{} is already powered down.".format(amc_num))
+            success = True
+        else:
+
+            # Sleep to allow previous telnet connection to properly
+            # close
+            time.sleep(1)
+
+            if self._session_open is False:
+                self.open()
+
+            self.logger.logger.info("Powering down AMC {}.".format(amc_num))
+            # Send shutdown command
+            self._send_command("shutdown {}".format(fru_num))
+
+            # Sleep to allow device in AMC slot to power down
+            self.logger.logger.info("Sleeping for {} seconds.".format(sleep))
+            time.sleep(sleep)
+
+            # Get updated state
+            _, state = self._get_slot_state(amc_num)
+
+            if state == "M1":
+                success = True
+                self.logger.logger.info(
+                    "AMC {} powered down successfully.".format(amc_num)
+                )
+            else:
+                self.logger.logger.warning(
+                    "Failed to power down slot. Current state: {}.".format(state)
+                )
+
+        self.close()
+
+        return success
+
+    def _slot_status_all(self):
+        """Internal method to get the current state of all
+        AMC slots.
+
+        Args:
+            close_conn(bool): Whether to close the telnet
+               session when returning, or not.
+               Default: True
+
+        Returns:
+            A list of tuples containing the AMC number and state code {M1, M2, ... M7} for all slots.
+            If a slot is empty, the entry in the list will contain None.
+        """
+        self.logger.logger.debug("Getting state of AMCs in chassis.")
+
+        if self._session_open is False:
+            self.open()
+
+        # Get fru_info for all AMCs
+        self._send_command("show_fru")
+        fru_info = self._read_command()
+
+        self.close()
+
+        # Search for populated AMC slots
+        match = re.findall(r"AMC(\d+)(\s+)(M\d+)", fru_info)
+
+        amcs = []
+
+        # Create list for all AMCs
+        for x in range(1, int(match[-1][0]) + 1):
+            slot_state = [item for item in match if str(x) in item]
+            if slot_state:
+                self.logger.logger.debug(
+                    "AMC{} slot is populated, and in the {} state.".format(
+                        x, slot_state[0][2]
+                    )
+                )
+                amcs.append((x, slot_state[0][2]))
+            else:
+                self.logger.logger.debug("AMC{} slot is unpopulated.".format(x))
+                amcs.append((x, None))
+
+        return amcs
+
+    def _stop_all_slots(self):
+        """Internal method to power down all AMC slots"""
+
+        if self._session_open is False:
+            self.open()
+
+        self.logger.logger.info("Powering down all AMCs.")
+
+        # Shutdown all slots
+        self._send_command("shutdown all")
+
+        self.logger.logger.info("All AMCs are now powered down.")
+
+        self.close()
+
+    def _determine_upstream(self):
+        """Internal method to determine which AMCs are configured
+        as 'Upstream', i.e the PCIe root-complex
+        """
+        if self._session_open is False:
+            self.open()
+
+        self.logger.logger.info("Determining which AMC(s) are configured as Upstream")
+
+        # 'mch' command takes a while to return all info
+        # so add an extended sleep period
+        self._send_command("mch", sleep=2)
+        mch_info = self._read_command()
+
+        self.close()
+
+        # Find the relevant lines in the output
+        first = mch_info.find("VS #")
+        last = mch_info.find("\r\nUpstream")
+        pcie_info = mch_info[first:last].splitlines()
+
+        """ Example info:
+            ['  VS #  | Host    | NT-Host | Members',
+             '  0       AMC01_4   none      AMC01_4 AMC02_4 AMC03_4 AMC04_4 AMC05_4 AMC06_4 AMC07_4 AMC08_4 AMC09_4 AMC10_4 AMC11_4 AMC12_4 ',
+             '  1     ',
+             '  2     ',
+             '  3     ',
+             '  4     ',
+             '  5     ']
+        """
+        upstreams = []
+        # There are upto 6  virtual switches
+        for i in range(1, len(pcie_info)):
+            # First line is the header
+            match = re.findall(r"{}(\s+)AMC(\d+)_(\d)".format(i - 1), pcie_info[i])
+            if len(match) > 0:
+                self.logger.logger.info(
+                    "AMC{} is the Upstream for Virtual Switch {}".format(
+                        int(match[0][1]), i - 1
+                    )
+                )
+                upstreams.append(int(match[0][1]))
+
+        return upstreams
+
+    def reboot_slot(self, amc_num):
+        """ Method to power cycle an individual AMC slot.
+
+        Args:
+            amc_num: AMC slot number to power cycle.
+
+        Returns:
+            A tuple containing the a success flag in the first position. In the \
+            second position:
+            - On success, an empty string.
+            - On failure, a string containing the failure message.
+        """
+        self.logger.logger.info("Rebooting AMC in slot {}".format(amc_num))
+
+        response = ""
+
+        # Sleep time
+        shutdown_sleep = 5
+        start_sleep = 5
+
+        # Shutdown slot
+        success = self._stop_slot(amc_num, shutdown_sleep)
+        if success is False:
+            response = "Failed to power down card in AMC slot {}.".format(amc_num)
+            self.logger.logger.warning(response)
+            return (success, response)
+
+        # Allow time for session to update
+        time.sleep(1)
+
+        # Start slot
+        success = self._start_slot(amc_num, start_sleep)
+        if success is False:
+            response = "Failed to power up card in AMC slot {}.".format(amc_num)
+            self.logger.logger.warning(response)
+
+        return (success, response)
+
+    def reboot_all_slots(self):
+        """ Method to power cycle all AMC slots.
+
+        Returns:
+            A tuple containing the a success flag in the first position. In the \
+            second position:
+            - On success, an empty string.
+            - On failure, a string containing the failure message.
+        """
+
+        if self._session_open is False:
+            self.open()
+
+        # Find which slots are populated
+        slot_status = self._slot_status_all()
+
+        # Shutdown all slots
+        self._stop_all_slots()
+
+        # Determine what AMC(s) are the Upstream (root-complex)
+        upstreams = self._determine_upstream()
+        upstream_state = []
+
+        overall_success = True
+        overall_response = ""
+        # Power on relevant slots
+        for slot in slot_status:
+            if (slot[1] is not None) and (slot[1] != "M1"):
+                if slot[0] not in upstreams:
+                    success = self._start_slot(slot[0])
+                    if success is False:
+                        response = "Failed to power up card in AMC slot {}.".format(
+                            slot[0]
+                        )
+                        self.logger.logger.warning(response)
+                        overall_success = False
+                        overall_response += "{}\n".format(response)
+                else:
+                    upstream_state.append(slot)
+                    self.logger.logger.info(
+                        "AMC{} is configured as an upstream. Will reboot last.".format(
+                            slot[0]
+                        )
+                    )
+
+        # Power on Upstream slots
+        for slot in upstream_state:
+            if (slot[1] is not None) and (slot[1] != "M1"):
+                success = self._start_slot(slot[0], upstream=True)
+                if success is False:
+                    response = "Failed to power up card in AMC slot {}.".format(slot[0])
+                    self.logger.logger.warning(response)
+                    overall_success = False
+                    overall_response += "{}\n".format(response)
+
+        self.close()
+
+        return (overall_success, overall_response)
