@@ -21,17 +21,17 @@ from __future__ import annotations
 import re
 import time
 import requests as rq
-from logging import Logger
 from collections import OrderedDict
 from bs4 import BeautifulSoup
 from ..gendev_err import FeatureNotSupported, NoRouteToDevice, WebChanged
+from ..gendev_logger import GenDevLogger
 
 __author__ = ["Felipe Torres González", "Ross Elliot"]
 __copyright__ = "Copyright 2021, ESS MCH Tools"
 __license__ = "GPL-3.0"
-__version__ = "1.0"
-__maintainer__ = "Felipe Torres González"
-__email__ = "felipe.torresgonzalez@ess.eu"
+__version__ = "1.1"
+__maintainer__ = "Ross Elliot"
+__email__ = "ross.elliot@ess.eu"
 __status__ = "Release Candidate"
 
 
@@ -55,7 +55,13 @@ class NATMCHWeb:
     - Check the backplane configuration of the mch against the content of a file.
     """
 
-    def __init__(self, ip_address: str, logger: Logger = None, open: bool = True):
+    def __init__(
+        self,
+        ip_address: str,
+        logger: GenDevLogger = None,
+        open: bool = True,
+        **kwargs,
+    ):
         """Class constructor.
 
         Args:
@@ -71,6 +77,11 @@ class NATMCHWeb:
         """
         self.ip_address = ip_address
 
+        if "log_to_file" in kwargs:
+            self._log_to_file = kwargs["log_to_file"]
+        else:
+            self._log_to_file = False
+
         # Header for the HTML methods, the most important variable is the
         # Authorization because NAT MCHs need to login using Root:NAT.
         self._http_headers = {
@@ -79,6 +90,17 @@ class NATMCHWeb:
             "Connection": "keep-alive",
         }
 
+        # Initialise Logger instance
+        if isinstance(logger, (GenDevLogger)):
+            self.logger = logger
+        else:
+            self.logger = GenDevLogger(logger=logger, log_to_file=self._log_to_file)
+
+        self.logger.logger.info(
+            "NATMCHWeb - NAT MCH Web instance created."
+            "\tMCH IP Address: {}".format(self.ip_address)
+        )
+
         # Flag indicating when the link is actually established
         #
         self._link_established = open
@@ -125,7 +147,7 @@ class NATMCHWeb:
         }
 
     def open(self):
-        """Stablish the connection to the target MCH.
+        """Establish the connection to the target MCH.
 
         Raises:
             NoRouteToDevice: when the device is not reachable in the network, or the \
@@ -139,28 +161,28 @@ class NATMCHWeb:
             )
         except Exception as e:
             if isinstance(e, rq.exceptions.ConnectionError):
-                raise NoRouteToDevice(
-                    "Error connecting to MCH web interface at {0}:".format(
-                        self.ip_address
-                    )
+                msg = "Error connecting to MCH web interface at {0}:".format(
+                    self.ip_address
                 )
+                self.logger.logger.error(msg)
+                raise NoRouteToDevice(msg)
 
         if not response.ok:
-            raise NoRouteToDevice(
-                "Unable to reach device at {0}, status code {1}:".format(
-                    self.ip_address, response.status_code
-                )
+            msg = "Unable to reach device at {0}, status code {1}:".format(
+                self.ip_address, response.status_code
             )
+            self.logger.logger.error(msg)
+            raise NoRouteToDevice(msg)
         else:
             html_content = BeautifulSoup(response.text, "lxml")
             title = html_content.head.title.text
 
             if not title == "MCH Configuration":
-                raise NoRouteToDevice(
-                    "Device at IP {0} is not an MCH. Retrieved description of device: ''{1}''".format(
-                        self.ip_address, title
-                    )
+                msg = "Device at IP {0} is not an MCH. Retrieved description of device: ''{1}''".format(
+                    self.ip_address, title
                 )
+                self.logger.logger.error(msg)
+                raise NoRouteToDevice(msg)
 
         self._link_established = True
 
@@ -400,7 +422,7 @@ class NATMCHWeb:
 
         return outdict
 
-    def _reboot(self):
+    def _reboot(self, sleep: int = 60):
         """Reboot the MCH
 
         This internal method issues an HTTP POST request to the web page to perform
@@ -412,6 +434,10 @@ class NATMCHWeb:
         situation might happen because a slow response of the DCHP server, or because
         something broke the MCH's configuration, so it fails to boot.
 
+        Args:
+            sleep: indicates how many seconds to wait after returning from the
+            method. Write a 0 to avoid it.
+
         Returns:
             True if the MCH is accessible after issuing the reboot request and
             waiting for a timeout.
@@ -420,6 +446,8 @@ class NATMCHWeb:
         Raises:
             NoRouteToDevice when the MCH couldn't be accessed after the reboot.
         """
+        self.logger.logger.info("Rebooting MCH. Sleeping for {} seconds.".format(sleep))
+
         data = {"save": "Reboot"}
         # The MCH doesn't return anything to the HTTP POST for the rebooting page
         # Use a timeout and check later that the MCH is up again
@@ -431,20 +459,18 @@ class NATMCHWeb:
                 timeout=1,
             )
         except rq.ReadTimeout:
-            # Add something to the log
             # Wait until the MCH is up
-            time.sleep(30)
+            time.sleep(sleep)
             # Sometimes, the MCH takes longer to reboot
             for i in range(10):
                 try:
-                    ret = self._check_is_mch()
-                    if ret:
-                        return True
+                    self.open()
+                    return True
                 except Exception:
                     if i == 9:
-                        raise NoRouteToDevice(
-                            "The MCH didn't come back from the reboot."
-                        )
+                        msg = "The MCH didn't come back from the reboot."
+                        self.logger.logger.error(msg)
+                        raise NoRouteToDevice(msg)
                     else:
                         time.sleep(10)
                         continue
@@ -536,9 +562,9 @@ class NATMCHWeb:
         I couldn't find a solution for that, which doesn't means it doesn't
         exists, because my experience with web programming is quite limited.
         """
-        raise FeatureNotSupported(
-            "The firmware update is not supported by this module."
-        )
+        msg = "The firmware update is not supported by this module."
+        self.logger.logger.info(msg)
+        raise FeatureNotSupported(msg)
 
     def set_configuration(
         self, category: str, data: OrderedDict, apply: bool = True
@@ -576,15 +602,16 @@ class NATMCHWeb:
             NoRouteToDevice: when the device is not reachable after applying the \
             new configuration.
         """
+        self.logger.logger.info("Setting {} configuration".format(category))
+
         if not self._link_established:
             self.open()
 
         validargs, _ = self._check_arguments(data, category)
         if not validargs:
-            return (
-                False,
-                "The given data doesn't match the expectation.",
-            )
+            msg = "The given data doesn't match the expectation."
+            self.logger.logger.error(msg)
+            return (False, msg)
 
         # Check the input parameter
         if category == "basecfg":
@@ -614,19 +641,20 @@ class NATMCHWeb:
                 {},
             ]
         elif category == "backplane":
-            raise FeatureNotSupported(
+            msg = (
                 "Changing the backplane configuration is not supported by this module."
             )
+            self.logger.logger.error(msg)
+            raise FeatureNotSupported(msg)
         else:
             cfgword = []
 
         if cfgword == []:
-            return (
-                False,
-                "The given category ({}) is not valid. See the documentation.".format(
-                    category
-                ),
+            msg = "The given category ({}) is not valid. See the documentation.".format(
+                category
             )
+            self.logger.logger.error(msg)
+            return (False, msg)
 
         # And this key is needed at the end of the dictionary
         saveword = "Save"
@@ -648,13 +676,15 @@ class NATMCHWeb:
 
         if response.ok:
             return_value = True, ""
-        else:
-            return_value = (
-                False,
-                "Configuration update for '{}' failed. Check the logs.".format(
-                    category
-                ),
+            self.logger.logger.info(
+                "Setting configuration of {} succeeded.".format(category)
             )
+        else:
+            msg = "Configuration update for '{}' failed. Check the logs.".format(
+                category
+            )
+            self.logger.logger.error(msg)
+            return_value = (False, msg)
 
         if apply:
             # This call might raise NoRouteToDevice, it should be handled upstream
@@ -690,6 +720,8 @@ class NATMCHWeb:
             config file*) and the whole configuration file as value for that \
             key.
         """
+        self.logger.logger.info("Getting {} configuration".format(category))
+
         if not self._link_established:
             self.open()
 
@@ -707,6 +739,11 @@ class NATMCHWeb:
         mch_config: OrderedDict = OrderedDict()
 
         if cfgword == "":
+            self.logger.logger.error(
+                'Get configuration failed for unsupported category "{}".'.format(
+                    category
+                )
+            )
             return False, mch_config
 
         response = rq.get(
@@ -755,27 +792,26 @@ class NATMCHWeb:
             checking was successful or not; and a text message when the
             checking was unsuccessful.
         """
+        self.logger.logger.info("Checking {} configuration".format(category))
+
         if not self._link_established:
             self.open()
 
         validargs, _ = self._check_arguments(config, category)
         if not validargs:
-            return (
-                False,
-                "The given data doesn't match the expectation.",
-            )
+            msg = "The given data doesn't match the expectation."
+            self.logger.logger.error(msg)
+            return (False, msg)
 
         # Check the input parameter
         if category == "backplane":
             cfgword = self._cfgkeys[category]["webkey"][0]
             cfgkey = self._cfgkeys[category]["cfgkey"]
         else:
-            return (
-                False,
-                "The given category ({}) is not valid. See the documentation.".format(
-                    category
-                ),
+            msg = "The given category ({}) is not valid. See the documentation.".format(
+                category
             )
+            return (False, msg)
 
         response = rq.post(
             "http://{}/goform/{}".format(self.ip_address, cfgword),
@@ -788,6 +824,10 @@ class NATMCHWeb:
             validation_result = htmlsoup.find("div", {"class": "info"}).text
 
             if "PASSED" in validation_result:
+                msg = "Configuration check for {} passed.\n{}".format(
+                    category, validation_result
+                )
+                self.logger.logger.info(msg)
                 return (True, validation_result)
         else:
             validation_result = "Bad request"
